@@ -45,20 +45,25 @@ impl super::Adapter {
         instance_flags: crate::InstanceFlags,
     ) -> Option<crate::ExposedAdapter<super::Api>> {
         // Create the device so that we can get the capabilities.
-        let device = match library.create_device(adapter, native::FeatureLevel::L11_0) {
-            Ok(pair) => match pair.into_result() {
-                Ok(device) => device,
+        let device = {
+            profiling::scope!("ID3D12Device::create_device");
+            match library.create_device(adapter, native::FeatureLevel::L11_0) {
+                Ok(pair) => match pair.into_result() {
+                    Ok(device) => device,
+                    Err(err) => {
+                        log::warn!("Device creation failed: {}", err);
+                        return None;
+                    }
+                },
                 Err(err) => {
-                    log::warn!("Device creation failed: {}", err);
+                    log::warn!("Device creation function is not found: {:?}", err);
                     return None;
                 }
-            },
-            Err(err) => {
-                log::warn!("Device creation function is not found: {:?}", err);
-                return None;
             }
         };
 
+        profiling::scope!("feature queries");
+
         // We have found a possible adapter.
         // Acquire the device information.
         let mut desc: dxgi1_2::DXGI_ADAPTER_DESC2 = unsafe { mem::zeroed() };
@@ -174,10 +179,12 @@ impl super::Adapter {
             | wgt::Features::MULTI_DRAW_INDIRECT
             | wgt::Features::MULTI_DRAW_INDIRECT_COUNT
             | wgt::Features::ADDRESS_MODE_CLAMP_TO_BORDER
-            | wgt::Features::NON_FILL_POLYGON_MODE
+            | wgt::Features::POLYGON_MODE_LINE
+            | wgt::Features::POLYGON_MODE_POINT
             | wgt::Features::VERTEX_WRITABLE_STORAGE
             | wgt::Features::TIMESTAMP_QUERY
-            | wgt::Features::TEXTURE_COMPRESSION_BC;
+            | wgt::Features::TEXTURE_COMPRESSION_BC
+            | wgt::Features::CLEAR_COMMANDS;
         //TODO: in order to expose this, we need to run a compute shader
         // that extract the necessary statistics out of the D3D12 result.
         // Alternatively, we could allocate a buffer for the query set,
@@ -236,6 +243,9 @@ impl super::Adapter {
                     max_vertex_attributes: d3d12::D3D12_IA_VERTEX_INPUT_RESOURCE_SLOT_COUNT,
                     max_vertex_buffer_array_stride: d3d12::D3D12_SO_BUFFER_MAX_STRIDE_IN_BYTES,
                     max_push_constant_size: 0,
+                    min_uniform_buffer_offset_alignment:
+                        d3d12::D3D12_CONSTANT_BUFFER_DATA_PLACEMENT_ALIGNMENT,
+                    min_storage_buffer_offset_alignment: 4, // TODO?
                 },
                 alignments: crate::Alignments {
                     buffer_copy_offset: wgt::BufferSize::new(
@@ -246,11 +256,6 @@ impl super::Adapter {
                         d3d12::D3D12_TEXTURE_DATA_PITCH_ALIGNMENT as u64,
                     )
                     .unwrap(),
-                    uniform_buffer_offset: wgt::BufferSize::new(
-                        d3d12::D3D12_CONSTANT_BUFFER_DATA_PLACEMENT_ALIGNMENT as u64,
-                    )
-                    .unwrap(),
-                    storage_buffer_offset: wgt::BufferSize::new(4).unwrap(), //TODO?
                 },
                 downlevel: wgt::DownlevelCapabilities::default(),
             },
@@ -262,16 +267,19 @@ impl crate::Adapter<super::Api> for super::Adapter {
     unsafe fn open(
         &self,
         features: wgt::Features,
+        _limits: &wgt::Limits,
     ) -> Result<crate::OpenDevice<super::Api>, crate::DeviceError> {
-        let queue = self
-            .device
-            .create_command_queue(
-                native::CmdListType::Direct,
-                native::Priority::Normal,
-                native::CommandQueueFlags::empty(),
-                0,
-            )
-            .into_device_result("Queue creation")?;
+        let queue = {
+            profiling::scope!("ID3D12Device::CreateCommandQueue");
+            self.device
+                .create_command_queue(
+                    native::CmdListType::Direct,
+                    native::Priority::Normal,
+                    native::CommandQueueFlags::empty(),
+                    0,
+                )
+                .into_device_result("Queue creation")?
+        };
 
         let device = super::Device::new(
             self.device,
