@@ -1,7 +1,10 @@
 use super::conv;
 use crate::auxil::map_naga_stage;
 use glow::HasContext;
-use std::{convert::TryInto, iter, mem, ptr, sync::Arc};
+use std::{convert::TryInto, iter, ptr, sync::Arc};
+
+#[cfg(not(target_arch = "wasm32"))]
+use std::mem;
 
 type ShaderStage<'a> = (
     naga::ShaderStage,
@@ -81,7 +84,7 @@ impl super::Device {
         gl: &glow::Context,
         shader: &str,
         naga_stage: naga::ShaderStage,
-        label: Option<&str>,
+        #[cfg_attr(target_arch = "wasm32", allow(unused))] label: Option<&str>,
     ) -> Result<glow::Shader, crate::PipelineError> {
         let target = match naga_stage {
             naga::ShaderStage::Vertex => glow::VERTEX_SHADER,
@@ -90,6 +93,7 @@ impl super::Device {
         };
 
         let raw = gl.create_shader(target).unwrap();
+        #[cfg(not(target_arch = "wasm32"))]
         if gl.supports_debug() {
             //TODO: remove all transmutes from `object_label`
             // https://github.com/grovesNL/glow/issues/186
@@ -170,9 +174,10 @@ impl super::Device {
         gl: &glow::Context,
         shaders: I,
         layout: &super::PipelineLayout,
-        label: crate::Label,
+        #[cfg_attr(target_arch = "wasm32", allow(unused))] label: Option<&str>,
     ) -> Result<super::PipelineInner, crate::PipelineError> {
         let program = gl.create_program().unwrap();
+        #[cfg(not(target_arch = "wasm32"))]
         if let Some(label) = label {
             if gl.supports_debug() {
                 gl.object_label(glow::PROGRAM, mem::transmute(program), Some(label));
@@ -325,26 +330,46 @@ impl crate::Device<super::Api> for super::Device {
             .contains(crate::MemoryFlags::PREFER_COHERENT);
         let mut map_flags = 0;
 
-        if is_host_visible {
-            map_flags |= glow::MAP_PERSISTENT_BIT;
-            if is_coherent {
-                map_flags |= glow::MAP_COHERENT_BIT;
-            }
-        }
-        if desc.usage.contains(crate::BufferUses::MAP_READ) {
-            map_flags |= glow::MAP_READ_BIT;
-        }
-        if desc.usage.contains(crate::BufferUses::MAP_WRITE) {
-            map_flags |= glow::MAP_WRITE_BIT;
-        }
-
         let raw = gl.create_buffer().unwrap();
         gl.bind_buffer(target, Some(raw));
         let raw_size = desc
             .size
             .try_into()
             .map_err(|_| crate::DeviceError::OutOfMemory)?;
-        gl.buffer_storage(target, raw_size, None, map_flags);
+
+        if self
+            .shared
+            .downlevel_flags
+            .contains(wgt::DownlevelFlags::VERTEX_STORAGE | wgt::DownlevelFlags::FRAGMENT_STORAGE)
+        {
+            if is_host_visible {
+                map_flags |= glow::MAP_PERSISTENT_BIT;
+                if is_coherent {
+                    map_flags |= glow::MAP_COHERENT_BIT;
+                }
+            }
+            if desc.usage.contains(crate::BufferUses::MAP_READ) {
+                map_flags |= glow::MAP_READ_BIT;
+            }
+            if desc.usage.contains(crate::BufferUses::MAP_WRITE) {
+                map_flags |= glow::MAP_WRITE_BIT;
+            }
+
+            gl.buffer_storage(target, raw_size, None, map_flags);
+        } else {
+            assert!(!is_coherent);
+            let usage = if is_host_visible {
+                if desc.usage.contains(crate::BufferUses::MAP_READ) {
+                    glow::STREAM_READ
+                } else {
+                    glow::DYNAMIC_DRAW
+                }
+            } else {
+                glow::STATIC_DRAW
+            };
+            gl.buffer_data_size(target, raw_size, usage);
+        }
+
         gl.bind_buffer(target, None);
 
         if !is_coherent && desc.usage.contains(crate::BufferUses::MAP_WRITE) {
@@ -352,6 +377,7 @@ impl crate::Device<super::Api> for super::Device {
         }
         //TODO: do we need `glow::MAP_UNSYNCHRONIZED_BIT`?
 
+        #[cfg(not(target_arch = "wasm32"))]
         if let Some(label) = desc.label {
             if gl.supports_debug() {
                 gl.object_label(glow::BUFFER, mem::transmute(raw), Some(label));
@@ -363,6 +389,7 @@ impl crate::Device<super::Api> for super::Device {
             target,
             size: desc.size,
             map_flags,
+            emulate_map_allocation: Default::default(),
         })
     }
     unsafe fn destroy_buffer(&self, buffer: super::Buffer) {
@@ -379,14 +406,28 @@ impl crate::Device<super::Api> for super::Device {
 
         let is_coherent = buffer.map_flags & glow::MAP_COHERENT_BIT != 0;
 
-        gl.bind_buffer(buffer.target, Some(buffer.raw));
-        let ptr = gl.map_buffer_range(
-            buffer.target,
-            range.start as i32,
-            (range.end - range.start) as i32,
-            buffer.map_flags,
-        );
-        gl.bind_buffer(buffer.target, None);
+        let ptr = if self
+            .shared
+            .workarounds
+            .contains(super::Workarounds::EMULATE_BUFFER_MAP)
+        {
+            let mut buf = vec![0; buffer.size as usize];
+            let ptr = buf.as_mut_ptr();
+            *buffer.emulate_map_allocation.lock().unwrap() = Some(buf);
+
+            ptr
+        } else {
+            gl.bind_buffer(buffer.target, Some(buffer.raw));
+            let ptr = gl.map_buffer_range(
+                buffer.target,
+                range.start as i32,
+                (range.end - range.start) as i32,
+                buffer.map_flags,
+            );
+            gl.bind_buffer(buffer.target, None);
+
+            ptr
+        };
 
         Ok(crate::BufferMapping {
             ptr: ptr::NonNull::new(ptr).ok_or(crate::DeviceError::Lost)?,
@@ -396,7 +437,14 @@ impl crate::Device<super::Api> for super::Device {
     unsafe fn unmap_buffer(&self, buffer: &super::Buffer) -> Result<(), crate::DeviceError> {
         let gl = &self.shared.context.lock();
         gl.bind_buffer(buffer.target, Some(buffer.raw));
-        gl.unmap_buffer(buffer.target);
+
+        if let Some(buf) = buffer.emulate_map_allocation.lock().unwrap().take() {
+            gl.buffer_sub_data_u8_slice(buffer.target, 0, &buf);
+            drop(buf);
+        } else {
+            gl.unmap_buffer(buffer.target);
+        }
+
         gl.bind_buffer(buffer.target, None);
         Ok(())
     }
@@ -407,11 +455,15 @@ impl crate::Device<super::Api> for super::Device {
         let gl = &self.shared.context.lock();
         gl.bind_buffer(buffer.target, Some(buffer.raw));
         for range in ranges {
-            gl.flush_mapped_buffer_range(
-                buffer.target,
-                range.start as i32,
-                (range.end - range.start) as i32,
-            );
+            if let Some(buf) = buffer.emulate_map_allocation.lock().unwrap().as_ref() {
+                gl.buffer_sub_data_u8_slice(buffer.target, range.start as i32, buf);
+            } else {
+                gl.flush_mapped_buffer_range(
+                    buffer.target,
+                    range.start as i32,
+                    (range.end - range.start) as i32,
+                );
+            }
         }
     }
     unsafe fn invalidate_mapped_ranges<I>(&self, _buffer: &super::Buffer, _ranges: I) {
@@ -429,6 +481,12 @@ impl crate::Device<super::Api> for super::Device {
             | crate::TextureUses::DEPTH_STENCIL_READ;
         let format_desc = self.shared.describe_texture_format(desc.format);
 
+        let mut copy_size = crate::CopyExtent {
+            width: desc.size.width,
+            height: desc.size.height,
+            depth: 1,
+        };
+
         let inner = if render_usage.contains(desc.usage)
             && desc.dimension == wgt::TextureDimension::D2
             && desc.size.depth_or_array_layers == 1
@@ -452,6 +510,7 @@ impl crate::Device<super::Api> for super::Device {
                 );
             }
 
+            #[cfg(not(target_arch = "wasm32"))]
             if let Some(label) = desc.label {
                 if gl.supports_debug() {
                     gl.object_label(glow::RENDERBUFFER, mem::transmute(raw), Some(label));
@@ -516,6 +575,7 @@ impl crate::Device<super::Api> for super::Device {
                     }
                 }
                 wgt::TextureDimension::D3 => {
+                    copy_size.depth = desc.size.depth_or_array_layers;
                     let target = glow::TEXTURE_3D;
                     gl.bind_texture(target, Some(raw));
                     gl.tex_storage_3d(
@@ -530,6 +590,7 @@ impl crate::Device<super::Api> for super::Device {
                 }
             };
 
+            #[cfg(not(target_arch = "wasm32"))]
             if let Some(label) = desc.label {
                 if gl.supports_debug() {
                     gl.object_label(glow::TEXTURE, mem::transmute(raw), Some(label));
@@ -562,6 +623,7 @@ impl crate::Device<super::Api> for super::Device {
             },
             format: desc.format,
             format_desc,
+            copy_size,
         })
     }
     unsafe fn destroy_texture(&self, texture: super::Texture) {
@@ -664,6 +726,7 @@ impl crate::Device<super::Api> for super::Device {
             );
         }
 
+        #[cfg(not(target_arch = "wasm32"))]
         if let Some(label) = desc.label {
             if gl.supports_debug() {
                 gl.object_label(glow::SAMPLER, mem::transmute(raw), Some(label));
@@ -951,11 +1014,11 @@ impl crate::Device<super::Api> for super::Device {
         gl.delete_program(pipeline.inner.program);
     }
 
+    #[cfg_attr(target_arch = "wasm32", allow(unused))]
     unsafe fn create_query_set(
         &self,
         desc: &wgt::QuerySetDescriptor<crate::Label>,
     ) -> Result<super::QuerySet, crate::DeviceError> {
-        use std::fmt::Write;
         let gl = &self.shared.context.lock();
         let mut temp_string = String::new();
 
@@ -964,7 +1027,10 @@ impl crate::Device<super::Api> for super::Device {
             let query = gl
                 .create_query()
                 .map_err(|_| crate::DeviceError::OutOfMemory)?;
+            #[cfg(not(target_arch = "wasm32"))]
             if gl.supports_debug() {
+                use std::fmt::Write;
+
                 if let Some(label) = desc.label {
                     temp_string.clear();
                     let _ = write!(temp_string, "{}[{}]", label, i);
@@ -1004,6 +1070,7 @@ impl crate::Device<super::Api> for super::Device {
         &self,
         fence: &super::Fence,
     ) -> Result<crate::FenceValue, crate::DeviceError> {
+        #[cfg_attr(target_arch = "wasm32", allow(clippy::needless_borrow))]
         Ok(fence.get_latest(&self.shared.context.lock()))
     }
     unsafe fn wait(
@@ -1012,7 +1079,7 @@ impl crate::Device<super::Api> for super::Device {
         wait_value: crate::FenceValue,
         timeout_ms: u32,
     ) -> Result<bool, crate::DeviceError> {
-        if fence.last_completed < wait_value {
+        if cfg!(not(target_arch = "wasm32")) && fence.last_completed < wait_value {
             let gl = &self.shared.context.lock();
             let timeout_ns = (timeout_ms as u64 * 1_000_000).min(!0u32 as u64);
             let &(_, sync) = fence
@@ -1033,9 +1100,8 @@ impl crate::Device<super::Api> for super::Device {
     unsafe fn start_capture(&self) -> bool {
         #[cfg(feature = "renderdoc")]
         {
-            //Note: it doesn't look like the device pointer is used by RD
             self.render_doc
-                .start_frame_capture(ptr::null_mut(), ptr::null_mut())
+                .start_frame_capture(self.shared.context.egl_context.as_ptr(), ptr::null_mut())
         }
         #[cfg(not(feature = "renderdoc"))]
         false
@@ -1046,3 +1112,9 @@ impl crate::Device<super::Api> for super::Device {
             .end_frame_capture(ptr::null_mut(), ptr::null_mut())
     }
 }
+
+// SAFE: WASM doesn't have threads
+#[cfg(target_arch = "wasm32")]
+unsafe impl Sync for super::Device {}
+#[cfg(target_arch = "wasm32")]
+unsafe impl Send for super::Device {}
