@@ -160,7 +160,14 @@ fn choose_config(
 
         match egl.choose_first_config(display, &attributes) {
             Ok(Some(config)) => {
-                return Ok((config, tier_max >= 1));
+                if tier_max == 1 {
+                    log::warn!(
+                        "EGL says it can present to the window but not natively. {}. {}",
+                        "This has been confirmed to malfunction on Intel+NV laptops",
+                        "Therefore, we disable presentation entirely for this platform"
+                    );
+                }
+                return Ok((config, tier_max >= 2));
             }
             Ok(None) => {
                 log::warn!("No config found!");
@@ -236,7 +243,7 @@ pub struct AdapterContext {
     glow_context: Mutex<glow::Context>,
     egl: Arc<egl::DynamicInstance<egl::EGL1_4>>,
     egl_display: egl::Display,
-    egl_context: egl::Context,
+    pub(super) egl_context: egl::Context,
     egl_pbuffer: Option<egl::Surface>,
 }
 
@@ -317,6 +324,7 @@ impl AdapterContext {
 #[derive(Debug)]
 struct Inner {
     egl: Arc<egl::DynamicInstance<egl::EGL1_4>>,
+    #[allow(unused)]
     version: (i32, i32),
     supports_native_window: bool,
     display: egl::Display,
@@ -596,7 +604,7 @@ impl crate::Instance<super::Api> for Instance {
                     return Err(crate::InstanceError);
                 }
             }
-            #[cfg(not(any(target_os = "android", target_os = "macos")))]
+            #[cfg(not(any(target_os = "android", target_os = "macos", target_os = "solaris")))]
             Rwh::Wayland(handle) => {
                 /* Wayland displays are not sharable between surfaces so if the
                  * surface we receive from this handle is from a different
@@ -721,6 +729,7 @@ pub struct Swapchain {
     extent: wgt::Extent3d,
     format: wgt::TextureFormat,
     format_desc: super::TextureFormatDesc,
+    #[allow(unused)]
     sample_type: wgt::TextureSampleType,
 }
 
@@ -731,6 +740,7 @@ pub struct Surface {
     config: egl::Config,
     display: egl::Display,
     context: egl::Context,
+    #[allow(unused)]
     pbuffer: Option<egl::Surface>,
     pub(super) presentable: bool,
     raw_window_handle: RawWindowHandle,
@@ -844,7 +854,11 @@ impl crate::Surface<super::Api> for Surface {
                     }
                     #[cfg(target_os = "android")]
                     Rwh::Android(handle) => handle.a_native_window,
-                    #[cfg(not(any(target_os = "android", target_os = "macos")))]
+                    #[cfg(not(any(
+                        target_os = "android",
+                        target_os = "macos",
+                        target_os = "solaris"
+                    )))]
                     Rwh::Wayland(handle) => {
                         let library = self.wsi_library.as_ref().expect("unsupported window");
                         let wl_egl_window_create: libloading::Symbol<WlEglWindowCreateFun> =
