@@ -58,6 +58,8 @@ To address this, we invalidate the vertex buffers based on:
 
 #[cfg(not(target_arch = "wasm32"))]
 mod egl;
+#[cfg(target_arch = "wasm32")]
+mod web;
 
 mod adapter;
 mod command;
@@ -68,6 +70,9 @@ mod queue;
 #[cfg(not(target_arch = "wasm32"))]
 use self::egl::{AdapterContext, Instance, Surface};
 
+#[cfg(target_arch = "wasm32")]
+use self::web::{AdapterContext, Instance, Surface};
+
 use arrayvec::ArrayVec;
 
 use glow::HasContext;
@@ -122,6 +127,11 @@ bitflags::bitflags! {
         const MEMORY_BARRIERS = 1 << 2;
         /// Vertex buffer layouts separate from the data.
         const VERTEX_BUFFER_LAYOUT = 1 << 3;
+        /// Indicates that buffers used as ELEMENT_ARRAY_BUFFER may be created / initialized / used
+        /// as other targets, if not present they must not be mixed with other targets.
+        const INDEX_BUFFER_ROLE_CHANGE = 1 << 4;
+        /// Indicates that the device supports disabling draw buffers
+        const CAN_DISABLE_DRAW_BUFFER = 1 << 5;
     }
 }
 
@@ -135,6 +145,8 @@ bitflags::bitflags! {
         // (https://gitlab.freedesktop.org/mesa/mesa/-/merge_requests/4972/diffs?diff_id=75888#22f5d1004713c9bbf857988c7efb81631ab88f99_323_327)
         // seems to indicate all skylake models are effected.
         const MESA_I915_SRGB_SHADER_CLEAR = 1 << 0;
+        /// Buffer map must emulated becuase it is not supported natively
+        const EMULATE_BUFFER_MAP = 1 << 1;
     }
 }
 
@@ -163,6 +175,7 @@ struct TextureFormatDesc {
 struct AdapterShared {
     context: AdapterContext,
     private_caps: PrivateCapabilities,
+    downlevel_flags: wgt::DownlevelFlags,
     workarounds: Workarounds,
     shading_language_version: naga::back::glsl::Version,
 }
@@ -188,11 +201,12 @@ pub struct Queue {
     shader_clear_program: glow::Program,
     /// The uniform location of the color uniform in the shader clear program
     shader_clear_program_color_uniform_location: glow::UniformLocation,
-    /// Keep a reasonably large buffer filled with zeroes, so that we can implement `FillBuffer` of
+    /// Keep a reasonably large buffer filled with zeroes, so that we can implement `ClearBuffer` of
     /// zeroes by copying from it.
     zero_buffer: glow::Buffer,
     temp_query_results: Vec<u64>,
     draw_buffer_count: u8,
+    current_index_buffer: Option<glow::Buffer>,
 }
 
 #[derive(Debug)]
@@ -201,8 +215,15 @@ pub struct Buffer {
     target: BindTarget,
     size: wgt::BufferAddress,
     map_flags: u32,
+    emulate_map_allocation: std::sync::Mutex<Option<Vec<u8>>>,
 }
 
+// Safe: WASM doesn't have threads
+#[cfg(target_arch = "wasm32")]
+unsafe impl Sync for Buffer {}
+#[cfg(target_arch = "wasm32")]
+unsafe impl Send for Buffer {}
+
 #[derive(Clone, Debug)]
 enum TextureInner {
     Renderbuffer {
@@ -217,7 +238,9 @@ enum TextureInner {
 impl TextureInner {
     fn as_native(&self) -> (glow::Texture, BindTarget) {
         match *self {
-            Self::Renderbuffer { raw, .. } => panic!("Unexpected renderbuffer {:?}", raw),
+            Self::Renderbuffer { .. } => {
+                panic!("Unexpected renderbuffer");
+            }
             Self::Texture { raw, target } => (raw, target),
         }
     }
@@ -229,6 +252,7 @@ pub struct Texture {
     mip_level_count: u32,
     array_layer_count: u32,
     format: wgt::TextureFormat,
+    #[allow(unused)]
     format_desc: TextureFormatDesc,
     copy_size: crate::CopyExtent,
 }
@@ -345,6 +369,7 @@ struct VertexBufferDesc {
     stride: u32,
 }
 
+#[allow(unused)]
 #[derive(Clone)]
 struct UniformDesc {
     location: glow::UniformLocation,
@@ -398,10 +423,22 @@ pub struct RenderPipeline {
     stencil: Option<StencilState>,
 }
 
+// SAFE: WASM doesn't have threads
+#[cfg(target_arch = "wasm32")]
+unsafe impl Send for RenderPipeline {}
+#[cfg(target_arch = "wasm32")]
+unsafe impl Sync for RenderPipeline {}
+
 pub struct ComputePipeline {
     inner: PipelineInner,
 }
 
+// SAFE: WASM doesn't have threads
+#[cfg(target_arch = "wasm32")]
+unsafe impl Send for ComputePipeline {}
+#[cfg(target_arch = "wasm32")]
+unsafe impl Sync for ComputePipeline {}
+
 #[derive(Debug)]
 pub struct QuerySet {
     queries: Box<[glow::Query]>,
@@ -528,11 +565,16 @@ enum Command {
         indirect_buf: glow::Buffer,
         indirect_offset: wgt::BufferAddress,
     },
-    FillBuffer {
+    ClearBuffer {
         dst: glow::Buffer,
         dst_target: BindTarget,
         range: crate::MemoryRange,
-        value: u8,
+    },
+    #[allow(unused)] // yet unimplemented
+    ClearTexture {
+        dst: glow::Texture,
+        dst_target: BindTarget,
+        subresource_range: wgt::ImageSubresourceRange,
     },
     CopyBufferToBuffer {
         src: glow::Buffer,
@@ -550,6 +592,7 @@ enum Command {
     },
     CopyBufferToTexture {
         src: glow::Buffer,
+        #[allow(unused)]
         src_target: BindTarget,
         dst: glow::Texture,
         dst_target: BindTarget,
@@ -561,6 +604,7 @@ enum Command {
         src_target: BindTarget,
         src_format: wgt::TextureFormat,
         dst: glow::Buffer,
+        #[allow(unused)]
         dst_target: BindTarget,
         copy: crate::BufferTextureCopy,
     },
