@@ -4,6 +4,8 @@ use parking_lot::Mutex;
 
 use std::{sync::Arc, thread};
 
+const MAX_COMMAND_BUFFERS: u64 = 2048;
+
 unsafe impl Send for super::Adapter {}
 unsafe impl Sync for super::Adapter {}
 
@@ -17,8 +19,13 @@ impl crate::Adapter<super::Api> for super::Adapter {
     unsafe fn open(
         &self,
         features: wgt::Features,
+        _limits: &wgt::Limits,
     ) -> Result<crate::OpenDevice<super::Api>, crate::DeviceError> {
-        let queue = self.shared.device.lock().new_command_queue();
+        let queue = self
+            .shared
+            .device
+            .lock()
+            .new_command_queue_with_max_command_buffer_count(MAX_COMMAND_BUFFERS);
         Ok(crate::OpenDevice {
             device: super::Device {
                 shared: Arc::clone(&self.shared),
@@ -863,7 +870,9 @@ impl super::PrivateCapabilities {
             | F::TEXTURE_COMPRESSION_BC
             | F::MAPPABLE_PRIMARY_BUFFERS
             | F::VERTEX_WRITABLE_STORAGE
-            | F::TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES;
+            | F::TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES
+            | F::POLYGON_MODE_LINE
+            | F::CLEAR_COMMANDS;
 
         features.set(F::DEPTH_CLAMPING, self.supports_depth_clamping);
 
@@ -891,7 +900,6 @@ impl super::PrivateCapabilities {
     }
 
     pub fn capabilities(&self) -> crate::Capabilities {
-        let buffer_alignment = wgt::BufferSize::new(self.buffer_alignment).unwrap();
         let mut downlevel = wgt::DownlevelCapabilities::default();
         downlevel.flags.set(
             wgt::DownlevelFlags::CUBE_ARRAY_TEXTURES,
@@ -929,12 +937,12 @@ impl super::PrivateCapabilities {
                 max_vertex_attributes: base.max_vertex_attributes,
                 max_vertex_buffer_array_stride: base.max_vertex_buffer_array_stride,
                 max_push_constant_size: 0x1000,
+                min_uniform_buffer_offset_alignment: self.buffer_alignment as u32,
+                min_storage_buffer_offset_alignment: self.buffer_alignment as u32,
             },
             alignments: crate::Alignments {
-                buffer_copy_offset: buffer_alignment,
+                buffer_copy_offset: wgt::BufferSize::new(self.buffer_alignment).unwrap(),
                 buffer_copy_pitch: wgt::BufferSize::new(4).unwrap(),
-                uniform_buffer_offset: buffer_alignment,
-                storage_buffer_offset: buffer_alignment,
             },
             downlevel,
         }
