@@ -242,7 +242,9 @@ impl super::DeviceShared {
         for &raw in self.framebuffers.lock().values() {
             self.raw.destroy_framebuffer(raw, None);
         }
-        self.raw.destroy_device(None);
+        if self.handle_is_owned {
+            self.raw.destroy_device(None);
+        }
     }
 }
 
@@ -375,7 +377,12 @@ impl
             })
             .collect::<ArrayVec<_, 8>>();
 
-        let mut vk_flags = vk::DescriptorPoolCreateFlags::empty();
+        let mut vk_flags =
+            if flags.contains(gpu_descriptor::DescriptorPoolCreateFlags::UPDATE_AFTER_BIND) {
+                vk::DescriptorPoolCreateFlags::UPDATE_AFTER_BIND
+            } else {
+                vk::DescriptorPoolCreateFlags::empty()
+            };
         if flags.contains(gpu_descriptor::DescriptorPoolCreateFlags::FREE_DESCRIPTOR_SET) {
             vk_flags |= vk::DescriptorPoolCreateFlags::FREE_DESCRIPTOR_SET;
         }
@@ -458,6 +465,12 @@ impl
     }
 }
 
+struct CompiledStage {
+    create_info: vk::PipelineShaderStageCreateInfo,
+    _entry_point: CString,
+    temp_raw_module: Option<vk::ShaderModule>,
+}
+
 impl super::Device {
     pub(super) unsafe fn create_swapchain(
         &self,
@@ -465,6 +478,7 @@ impl super::Device {
         config: &crate::SurfaceConfiguration,
         provided_old_swapchain: Option<super::Swapchain>,
     ) -> Result<super::Swapchain, crate::SurfaceError> {
+        profiling::scope!("Device::create_swapchain");
         let functor = khr::Swapchain::new(&surface.instance.raw, &self.shared.raw);
 
         let old_swapchain = match provided_old_swapchain {
@@ -491,7 +505,10 @@ impl super::Device {
             .clipped(true)
             .old_swapchain(old_swapchain);
 
-        let result = functor.create_swapchain(&info, None);
+        let result = {
+            profiling::scope!("vkCreateSwapchainKHR");
+            functor.create_swapchain(&info, None)
+        };
 
         // doing this before bailing out with error
         if old_swapchain != vk::SwapchainKHR::null() {
@@ -532,21 +549,19 @@ impl super::Device {
         })
     }
 
-    /// The image handle and its memory are not handled by wgpu-hal
-    ///
     /// # Safety
     ///
-    /// - `vk_image` must be created respecting `desc` and `raw_flags`
-    /// - The application must manually destroy the texture handle. This can be done inside the
-    ///   `Drop` impl of `drop_guard`.
+    /// - `vk_image` must be created respecting `desc`
+    /// - If `drop_guard` is `Some`, the application must manually destroy the image handle. This
+    ///   can be done inside the `Drop` impl of `drop_guard`.
     pub unsafe fn texture_from_raw(
         vk_image: vk::Image,
         desc: &crate::TextureDescriptor,
-        drop_guard: super::DropGuard,
+        drop_guard: Option<super::DropGuard>,
     ) -> super::Texture {
         super::Texture {
             raw: vk_image,
-            drop_guard: Some(drop_guard),
+            drop_guard,
             block: None,
             usage: desc.usage,
             aspects: crate::FormatAspects::from(desc.format),
@@ -555,6 +570,82 @@ impl super::Device {
             copy_size: conv::map_extent_to_copy_size(&desc.size, desc.dimension),
         }
     }
+
+    fn create_shader_module_impl(
+        &self,
+        spv: &[u32],
+    ) -> Result<vk::ShaderModule, crate::DeviceError> {
+        let vk_info = vk::ShaderModuleCreateInfo::builder()
+            .flags(vk::ShaderModuleCreateFlags::empty())
+            .code(spv);
+
+        let raw = unsafe {
+            profiling::scope!("vkCreateShaderModule");
+            self.shared.raw.create_shader_module(&vk_info, None)?
+        };
+        Ok(raw)
+    }
+
+    fn compile_stage(
+        &self,
+        stage: &crate::ProgrammableStage<super::Api>,
+        naga_stage: naga::ShaderStage,
+    ) -> Result<CompiledStage, crate::PipelineError> {
+        let stage_flags = crate::auxil::map_naga_stage(naga_stage);
+        let vk_module = match *stage.module {
+            super::ShaderModule::Raw(raw) => raw,
+            super::ShaderModule::Intermediate {
+                ref naga_shader,
+                runtime_checks,
+            } => {
+                let pipeline_options = naga::back::spv::PipelineOptions {
+                    entry_point: stage.entry_point.to_string(),
+                    shader_stage: naga_stage,
+                };
+                let temp_options;
+                let options = if !runtime_checks {
+                    temp_options = naga::back::spv::Options {
+                        bounds_check_policies: naga::back::BoundsCheckPolicies {
+                            index: naga::back::BoundsCheckPolicy::Unchecked,
+                            buffer: naga::back::BoundsCheckPolicy::Unchecked,
+                            image: naga::back::BoundsCheckPolicy::Unchecked,
+                        },
+                        ..self.naga_options.clone()
+                    };
+                    &temp_options
+                } else {
+                    &self.naga_options
+                };
+                let spv = {
+                    profiling::scope!("naga::spv::write_vec");
+                    naga::back::spv::write_vec(
+                        &naga_shader.module,
+                        &naga_shader.info,
+                        options,
+                        Some(&pipeline_options),
+                    )
+                }
+                .map_err(|e| crate::PipelineError::Linkage(stage_flags, format!("{}", e)))?;
+                self.create_shader_module_impl(&spv)?
+            }
+        };
+
+        let entry_point = CString::new(stage.entry_point).unwrap();
+        let create_info = vk::PipelineShaderStageCreateInfo::builder()
+            .stage(conv::map_shader_stage(stage_flags))
+            .module(vk_module)
+            .name(&entry_point)
+            .build();
+
+        Ok(CompiledStage {
+            create_info,
+            _entry_point: entry_point,
+            temp_raw_module: match *stage.module {
+                super::ShaderModule::Raw(_) => None,
+                super::ShaderModule::Intermediate { .. } => Some(vk_module),
+            },
+        })
+    }
 }
 
 impl crate::Device<super::Api> for super::Device {
@@ -953,9 +1044,69 @@ impl crate::Device<super::Api> for super::Device {
             })
             .collect::<Vec<_>>();
 
-        let vk_info = vk::DescriptorSetLayoutCreateInfo::builder()
-            .flags(vk::DescriptorSetLayoutCreateFlags::empty())
-            .bindings(&vk_bindings);
+        let vk_info = vk::DescriptorSetLayoutCreateInfo::builder().bindings(&vk_bindings);
+
+        let mut binding_flag_info;
+        let binding_flag_vec;
+        let mut requires_update_after_bind = false;
+
+        let partially_bound = desc
+            .flags
+            .contains(crate::BindGroupLayoutFlags::PARTIALLY_BOUND);
+
+        let vk_info = if !self.shared.uab_types.is_empty() || partially_bound {
+            binding_flag_vec = desc
+                .entries
+                .iter()
+                .map(|entry| {
+                    let mut flags = vk::DescriptorBindingFlags::empty();
+
+                    if partially_bound && entry.count.is_some() {
+                        flags |= vk::DescriptorBindingFlags::PARTIALLY_BOUND;
+                    }
+
+                    let uab_type = match entry.ty {
+                        wgt::BindingType::Buffer {
+                            ty: wgt::BufferBindingType::Uniform,
+                            ..
+                        } => super::UpdateAfterBindTypes::UNIFORM_BUFFER,
+                        wgt::BindingType::Buffer {
+                            ty: wgt::BufferBindingType::Storage { .. },
+                            ..
+                        } => super::UpdateAfterBindTypes::STORAGE_BUFFER,
+                        wgt::BindingType::Texture { .. } => {
+                            super::UpdateAfterBindTypes::SAMPLED_TEXTURE
+                        }
+                        wgt::BindingType::StorageTexture { .. } => {
+                            super::UpdateAfterBindTypes::STORAGE_TEXTURE
+                        }
+                        _ => super::UpdateAfterBindTypes::empty(),
+                    };
+
+                    if !uab_type.is_empty() && self.shared.uab_types.contains(uab_type) {
+                        flags |= vk::DescriptorBindingFlags::UPDATE_AFTER_BIND;
+                        requires_update_after_bind = true;
+                    }
+
+                    flags
+                })
+                .collect::<Vec<_>>();
+
+            binding_flag_info = vk::DescriptorSetLayoutBindingFlagsCreateInfo::builder()
+                .binding_flags(&binding_flag_vec);
+
+            vk_info.push_next(&mut binding_flag_info)
+        } else {
+            vk_info
+        };
+
+        let dsl_create_flags = if requires_update_after_bind {
+            vk::DescriptorSetLayoutCreateFlags::UPDATE_AFTER_BIND_POOL
+        } else {
+            vk::DescriptorSetLayoutCreateFlags::empty()
+        };
+
+        let vk_info = vk_info.flags(dsl_create_flags);
 
         let raw = self
             .shared
@@ -971,6 +1122,7 @@ impl crate::Device<super::Api> for super::Device {
             raw,
             desc_count,
             types: types.into_boxed_slice(),
+            requires_update_after_bind,
         })
     }
     unsafe fn destroy_bind_group_layout(&self, bg_layout: super::BindGroupLayout) {
@@ -1004,7 +1156,10 @@ impl crate::Device<super::Api> for super::Device {
             .set_layouts(&vk_set_layouts)
             .push_constant_ranges(&vk_push_constant_ranges);
 
-        let raw = self.shared.raw.create_pipeline_layout(&vk_info, None)?;
+        let raw = {
+            profiling::scope!("vkCreatePipelineLayout");
+            self.shared.raw.create_pipeline_layout(&vk_info, None)?
+        };
 
         if let Some(label) = desc.label {
             self.shared
@@ -1026,7 +1181,11 @@ impl crate::Device<super::Api> for super::Device {
         let mut vk_sets = self.desc_allocator.lock().allocate(
             &*self.shared,
             &desc.layout.raw,
-            gpu_descriptor::DescriptorSetLayoutCreateFlags::empty(),
+            if desc.layout.requires_update_after_bind {
+                gpu_descriptor::DescriptorSetLayoutCreateFlags::UPDATE_AFTER_BIND
+            } else {
+                gpu_descriptor::DescriptorSetLayoutCreateFlags::empty()
+            },
             &desc.layout.desc_count,
             1,
         )?;
@@ -1063,7 +1222,7 @@ impl crate::Device<super::Api> for super::Device {
                 vk::DescriptorType::SAMPLED_IMAGE | vk::DescriptorType::STORAGE_IMAGE => {
                     let index = image_infos.len();
                     let start = entry.resource_index;
-                    let end = start + size;
+                    let end = start + entry.count;
                     image_infos.extend(desc.textures[start as usize..end as usize].iter().map(
                         |binding| {
                             let layout =
@@ -1082,7 +1241,7 @@ impl crate::Device<super::Api> for super::Device {
                 | vk::DescriptorType::STORAGE_BUFFER_DYNAMIC => {
                     let index = buffer_infos.len();
                     let start = entry.resource_index;
-                    let end = start + size;
+                    let end = start + entry.count;
                     buffer_infos.extend(desc.buffers[start as usize..end as usize].iter().map(
                         |binding| {
                             vk::DescriptorBufferInfo::builder()
@@ -1114,36 +1273,54 @@ impl crate::Device<super::Api> for super::Device {
         shader: crate::ShaderInput,
     ) -> Result<super::ShaderModule, crate::ShaderError> {
         let spv = match shader {
-            crate::ShaderInput::Naga(naga_shader) => Cow::Owned(
-                naga::back::spv::write_vec(
-                    &naga_shader.module,
-                    &naga_shader.info,
-                    &self.naga_options,
+            crate::ShaderInput::Naga(naga_shader) => {
+                if self
+                    .shared
+                    .workarounds
+                    .contains(super::Workarounds::SEPARATE_ENTRY_POINTS)
+                {
+                    return Ok(super::ShaderModule::Intermediate {
+                        naga_shader,
+                        runtime_checks: desc.runtime_checks,
+                    });
+                }
+                let mut naga_options = self.naga_options.clone();
+                if !desc.runtime_checks {
+                    naga_options.bounds_check_policies = naga::back::BoundsCheckPolicies {
+                        index: naga::back::BoundsCheckPolicy::Unchecked,
+                        buffer: naga::back::BoundsCheckPolicy::Unchecked,
+                        image: naga::back::BoundsCheckPolicy::Unchecked,
+                    };
+                }
+                Cow::Owned(
+                    naga::back::spv::write_vec(
+                        &naga_shader.module,
+                        &naga_shader.info,
+                        &naga_options,
+                        None,
+                    )
+                    .map_err(|e| crate::ShaderError::Compilation(format!("{}", e)))?,
                 )
-                .map_err(|e| crate::ShaderError::Compilation(format!("{}", e)))?,
-            ),
+            }
             crate::ShaderInput::SpirV(spv) => Cow::Borrowed(spv),
         };
 
-        let vk_info = vk::ShaderModuleCreateInfo::builder()
-            .flags(vk::ShaderModuleCreateFlags::empty())
-            .code(&spv);
-
-        let raw = self
-            .shared
-            .raw
-            .create_shader_module(&vk_info, None)
-            .map_err(crate::DeviceError::from)?;
+        let raw = self.create_shader_module_impl(&*spv)?;
 
         if let Some(label) = desc.label {
             self.shared
                 .set_object_name(vk::ObjectType::SHADER_MODULE, raw, label);
         }
 
-        Ok(super::ShaderModule { raw })
+        Ok(super::ShaderModule::Raw(raw))
     }
     unsafe fn destroy_shader_module(&self, module: super::ShaderModule) {
-        let _ = self.shared.raw.destroy_shader_module(module.raw, None);
+        match module {
+            super::ShaderModule::Raw(raw) => {
+                let _ = self.shared.raw.destroy_shader_module(raw, None);
+            }
+            super::ShaderModule::Intermediate { .. } => {}
+        }
     }
 
     unsafe fn create_render_pipeline(
@@ -1193,29 +1370,16 @@ impl crate::Device<super::Api> for super::Device {
             .primitive_restart_enable(desc.primitive.strip_index_format.is_some())
             .build();
 
-        let vs_entry_point;
-        {
-            let stage = &desc.vertex_stage;
-            vs_entry_point = CString::new(stage.entry_point).unwrap();
-            stages.push(
-                vk::PipelineShaderStageCreateInfo::builder()
-                    .stage(vk::ShaderStageFlags::VERTEX)
-                    .module(stage.module.raw)
-                    .name(&vs_entry_point)
-                    .build(),
-            );
-        }
-        let fs_entry_point;
-        if let Some(ref stage) = desc.fragment_stage {
-            fs_entry_point = CString::new(stage.entry_point).unwrap();
-            stages.push(
-                vk::PipelineShaderStageCreateInfo::builder()
-                    .stage(vk::ShaderStageFlags::FRAGMENT)
-                    .module(stage.module.raw)
-                    .name(&fs_entry_point)
-                    .build(),
-            );
-        }
+        let compiled_vs = self.compile_stage(&desc.vertex_stage, naga::ShaderStage::Vertex)?;
+        stages.push(compiled_vs.create_info);
+        let compiled_fs = match desc.fragment_stage {
+            Some(ref stage) => {
+                let compiled = self.compile_stage(stage, naga::ShaderStage::Fragment)?;
+                stages.push(compiled.create_info);
+                Some(compiled)
+            }
+            None => None,
+        };
 
         let mut vk_rasterization = vk::PipelineRasterizationStateCreateInfo::builder()
             .depth_clamp_enable(desc.primitive.clamp_depth)
@@ -1343,11 +1507,13 @@ impl crate::Device<super::Api> for super::Device {
                 .build()
         }];
 
-        let mut raw_vec = self
-            .shared
-            .raw
-            .create_graphics_pipelines(vk::PipelineCache::null(), &vk_infos, None)
-            .map_err(|(_, e)| crate::DeviceError::from(e))?;
+        let mut raw_vec = {
+            profiling::scope!("vkCreateGraphicsPipelines");
+            self.shared
+                .raw
+                .create_graphics_pipelines(vk::PipelineCache::null(), &vk_infos, None)
+                .map_err(|(_, e)| crate::DeviceError::from(e))?
+        };
 
         let raw = raw_vec.pop().unwrap();
         if let Some(label) = desc.label {
@@ -1355,6 +1521,17 @@ impl crate::Device<super::Api> for super::Device {
                 .set_object_name(vk::ObjectType::PIPELINE, raw, label);
         }
 
+        if let Some(raw_module) = compiled_vs.temp_raw_module {
+            self.shared.raw.destroy_shader_module(raw_module, None);
+        }
+        if let Some(CompiledStage {
+            temp_raw_module: Some(raw_module),
+            ..
+        }) = compiled_fs
+        {
+            self.shared.raw.destroy_shader_module(raw_module, None);
+        }
+
         Ok(super::RenderPipeline { raw })
     }
     unsafe fn destroy_render_pipeline(&self, pipeline: super::RenderPipeline) {
@@ -1365,25 +1542,22 @@ impl crate::Device<super::Api> for super::Device {
         &self,
         desc: &crate::ComputePipelineDescriptor<super::Api>,
     ) -> Result<super::ComputePipeline, crate::PipelineError> {
-        let cs_entry_point = CString::new(desc.stage.entry_point).unwrap();
-        let vk_stage = vk::PipelineShaderStageCreateInfo::builder()
-            .stage(vk::ShaderStageFlags::COMPUTE)
-            .module(desc.stage.module.raw)
-            .name(&cs_entry_point)
-            .build();
+        let compiled = self.compile_stage(&desc.stage, naga::ShaderStage::Compute)?;
 
         let vk_infos = [{
             vk::ComputePipelineCreateInfo::builder()
                 .layout(desc.layout.raw)
-                .stage(vk_stage)
+                .stage(compiled.create_info)
                 .build()
         }];
 
-        let mut raw_vec = self
-            .shared
-            .raw
-            .create_compute_pipelines(vk::PipelineCache::null(), &vk_infos, None)
-            .map_err(|(_, e)| crate::DeviceError::from(e))?;
+        let mut raw_vec = {
+            profiling::scope!("vkCreateComputePipelines");
+            self.shared
+                .raw
+                .create_compute_pipelines(vk::PipelineCache::null(), &vk_infos, None)
+                .map_err(|(_, e)| crate::DeviceError::from(e))?
+        };
 
         let raw = raw_vec.pop().unwrap();
         if let Some(label) = desc.label {
@@ -1391,6 +1565,10 @@ impl crate::Device<super::Api> for super::Device {
                 .set_object_name(vk::ObjectType::PIPELINE, raw, label);
         }
 
+        if let Some(raw_module) = compiled.temp_raw_module {
+            self.shared.raw.destroy_shader_module(raw_module, None);
+        }
+
         Ok(super::ComputePipeline { raw })
     }
     unsafe fn destroy_compute_pipeline(&self, pipeline: super::ComputePipeline) {
@@ -1535,20 +1713,27 @@ impl crate::Device<super::Api> for super::Device {
     unsafe fn start_capture(&self) -> bool {
         #[cfg(feature = "renderdoc")]
         {
-            self.render_doc.start_frame_capture(
-                ash::vk::Handle::as_raw(self.shared.raw.handle()) as *mut _,
-                ptr::null_mut(),
-            )
+            // Renderdoc requires us to give us the pointer that vkInstance _points to_.
+            let raw_vk_instance =
+                ash::vk::Handle::as_raw(self.shared.instance.raw.handle()) as *mut *mut _;
+            let raw_vk_instance_dispatch_table = *raw_vk_instance;
+            self.render_doc
+                .start_frame_capture(raw_vk_instance_dispatch_table, ptr::null_mut())
         }
         #[cfg(not(feature = "renderdoc"))]
         false
     }
     unsafe fn stop_capture(&self) {
         #[cfg(feature = "renderdoc")]
-        self.render_doc.end_frame_capture(
-            ash::vk::Handle::as_raw(self.shared.raw.handle()) as *mut _,
-            ptr::null_mut(),
-        )
+        {
+            // Renderdoc requires us to give us the pointer that vkInstance _points to_.
+            let raw_vk_instance =
+                ash::vk::Handle::as_raw(self.shared.instance.raw.handle()) as *mut *mut _;
+            let raw_vk_instance_dispatch_table = *raw_vk_instance;
+
+            self.render_doc
+                .end_frame_capture(raw_vk_instance_dispatch_table, ptr::null_mut())
+        }
     }
 }
 
